/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>
#include <limits.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"

#include "graph.h"


/****************************************************************************

   Declaration of private functions (prototypes)
*/

flag_t
write_knots(FILE *out, petrinet_t *p, subnet_t *s, knot_type_t type);

flag_t
write_edges(FILE *out, petrinet_t *p, subnet_t *s);

unsigned char *
prepare_name(unsigned char *b, xmlChar *n);

/****************************************************************************

   Declaration of private constants
*/

/* const xmlChar font_default[] = "Vera"; */
const char font_default[] = "Helvetica";
const char color_default[] = "black";

/****************************************************************************

 Definition of functions
*/

flag_t
write_knots(FILE *out, petrinet_t *p, subnet_t *s, knot_type_t type)
{
  /*

  write_knots() - 

  */
  knot_t *c;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out,
	   "//\n// %s\n//\n",
	   ((type == state) ? "States" : "Transitions"));

  for (c=((type == state) ? p->s : p->t); c->id; c++) {	/* all knots w/ id */
    if (c->valid && c->out
	&& (!s || c->sn == s)) {

      fprintf( out,
	       "\"%s\" [\n"
	      "\tlabel = \"%s\"\n"
	      "\tfontsize = \"8\"\n"
	      "\tfixedsize= \"true\"\n"
	      "\tshape = \"%s\"\n"
	      "\tcolor = \"%s\"\n"
	      "\tperipheries = \"%s\"\n"
	      "\tURL = \"%s.html\"\n"
	      "];\n",
	      c->id,
/* 	      prepare_name(c->name), */
/* 	       prepare_name(buffer_name, c->name), */
	       ((c->name && strlen(c->name))
		? prepare_name(buffer_name, c->name)
		: c->id),
	      ((type == state) ? "circle" : "box"),
/* 	       ((c == p->start || c == p->target) ? "red" : "black"), */
	       (c->color ? c->color : color_default),
	       (c->rating ? c->rating : "1"),
/* 	       isolat1_str(doc_prefix), */
/* 	       p->doc_prefix, */
	       c->id);
    }
  }

  return true;
}
/* End of write_knots() */


/*! If 's' is the first subnet (== root), write all edges between
  different subnets _and_ the edges inside 's' itself.

*/
flag_t
write_edges(FILE *out, petrinet_t *p, subnet_t *s)
{

  index_t i_e;
  edge_t *c;

  fprintf( out, "//\n// Edges\n//\n");

#if 1
  for (c=p->edge; c->from != NULL; c++) {

    if ( c->valid && c->from->out && c->to->out
	 && (s == NULL
	     || ((s == p->sn &&
		  ((c->from->sn == p->sn && c->to->sn == p->sn) || c->from->sn != c->to->sn)
		  ) /* edge inside root subnet or between two subnets */
		 || (c->from->sn == s && c->to->sn == s))
	     )
	 ) {

/*       printf_debug( "%s -> %s", knotid(c->from), knotid(c->to)); */

      fprintf( out,
	       "\t\"%s\" -> \"%s\" [style = \"%s\", color = \"%s\"];\n",
	       knotid(c->from),
	       knotid(c->to),
	       ((c->style) ? c->style : "solid"),
	       ((c->color) ? c->color : "black")
	       );
    }
  }
#endif

#if 0
  for (i_e=0; i_e < p->n_e; i_e++) {
    if (p->edge[i_e].valid) {
/* 	assert(knotid(p->edge[i_e].from)); */
/*       assert(p->edge[i_e].from->valid);  */
/*       assert(p->edge[i_e].to->valid); */
      printf_debug( "%i %s -> %s", i_e, knotid(p->edge[i_e].from), knotid(p->edge[i_e].to));
/*       printf_debug( "%i %s -> %s", i_e, "f", "t"); */
    }
    else {
      printf_debug( "%i invalid", i_e);
    }
  }
#endif

#if 0
  for (i_e=0; i_e < p->n_e; i_e++) {
    if (p->edge[i_e].valid) {
      fprintf( stderr, "%s\n", p->edge[i_e].from->id);

      fprintf( out,
	       "\t\"%s\" -> \"%s\" [style = %s];\n",
	       (p->edge[i_e].from)->id, (p->edge[i_e].to)->id,
	       ((p->edge[i_e].style) ? p->edge[i_e].style : "normal"));
    }
  }
#endif

  return true;
}
/* End of write_edges() */


/*! insert a line break at '-' chars and use isolat1_str()
*/
unsigned char *
prepare_name(unsigned char *b, xmlChar *n)
{
  unsigned char buffer[BUFFER_LENGTH];

  int i;
  int j;

  isolat1_str(buffer,n,BUFFER_LENGTH);

  for (i=0, j=0; buffer[i] && j<1024; i++, j++) {

    b[j] = buffer[i];
    if (b[j] == '-' || b[j] == ' ') {
      b[++j] = '\\';
      b[++j] = 'n';
    }
  }
  b[j] = '\0';

  return b;
}
/* End of prepare_name() */


/*!

  's' == NULL -> no subgraphs

*/
flag_t
graph_subnet(FILE *out, petrinet_t *p, subnet_t *s, index_t number)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  if (!s || s->valid) {

    if (s) {
      fprintf( out,
	       "\tsubgraph cluster%i {\n"
	       "\t\tlabel = \"%s\"\n"
	       "\n",
	       number,
	       ((s->name && strlen(s->name))
		? prepare_name(buffer_name, s->name)
		: s->id));
    }

    write_knots( out, p, s, state);

    write_knots( out, p, s, transition);

    write_edges(out, p, s);

    if (s) {
      fprintf( out,
	       "\t} // end of subgraph '%s'\n\n",
	       isolat1_str(buffer_name, s->id, BUFFER_LENGTH));
    }
  }

  return true;
}
/* End of graph_subnet() */


/*! writes a html table of all path's

\param out Output file handle
\param p pointer to used petrinet
\param doc_prefix for link

\result flag only
 */
flag_t
graph_pathnet(FILE *out, petrinet_t *p)
{
  petrinet_pathnet_out_flag(p,true);

  graph_subnet(out,p,NULL,0);

  return (p->n_paths > 0);
}
/* End of graph_pathnet() */



/*! calls graphiz dot
*/
flag_t
graph_dot_call(xmlChar *id, char *doc_prefix)
{
  char graph_name[BUFFER_LENGTH];
  char buffer_cmd[BUFFER_LENGTH];

  assert(id);

  sprintf(graph_name, "%sgraph_%s", doc_prefix, (char *) id);

  /* dot create MAP file */
  sprintf(buffer_cmd, 
	  "dot -Tcmap -o %s.map %s.dot", 
	  graph_name, 
	  graph_name);

  printf_debug("Call '%s'", buffer_cmd);
  system(buffer_cmd);

  /* dot create PNG file */
  sprintf(buffer_cmd, 
	  "dot -Tpng -o %s.png %s.dot", 
	  graph_name, 
	  graph_name);

  printf_debug("Call '%s'", buffer_cmd);
  system(buffer_cmd);

  return true;
}
/* end of graph_dot_call() */


flag_t
graph_knot_file(petrinet_t *p, knot_t *k)
{
  /* graph_knot_file() - 
  */

  assert(k);

  if (k->valid && k->sn->valid) {

    char graph_name[BUFFER_LENGTH];
    FILE *filep_knot_graph;

    sprintf(graph_name, "%sgraph_%s.dot", p->doc_prefix, (char *) k->id);

    filep_knot_graph = fopen(graph_name, "w");

    printf_exit( (! filep_knot_graph), 2, "Cant create '%s'", graph_name );

    printf_debug("Write '%s'", graph_name);

    graph_header(filep_knot_graph, p);

    graph_knot(filep_knot_graph, p, k);

    graph_footer(filep_knot_graph);

    fclose(filep_knot_graph);

    return true;
  }

  return false;
}
/* End of graph_knot_file() */


flag_t
graph_subnet_file(petrinet_t *p, subnet_t *s, index_t number)
{
  /* graph_subnet_file() - 
  */

  assert(s);

  if (s->valid) {

    char graph_name[BUFFER_LENGTH];
    FILE *filep_subnet_graph;

    sprintf(graph_name, "%sgraph_%s.dot", p->doc_prefix, (char *) s->id);

    filep_subnet_graph = fopen(graph_name, "w");

    printf_exit( (! filep_subnet_graph), 2, "Cant create '%s'", graph_name );

    printf_debug("Write '%s'", graph_name);

    graph_header(filep_subnet_graph, p);

    graph_subnet(filep_subnet_graph, p, s, number);

    graph_footer(filep_subnet_graph);

    fclose(filep_subnet_graph);

    return true;
  }

  return false;
}
/* End of graph_subnet_file() */


flag_t
graph_knot(FILE *out, petrinet_t *p, knot_t *k)
{
  /*

  graph_knot() - 

  */
  unsigned char buffer_name[BUFFER_LENGTH];

  if (k->valid) {

    petrinet_all_out_flag(p, false);

    petrinet_neighbors_out_flag(p, k, true);

    fprintf( out,
	     "\tsubgraph %s {\n"
	     "\t\tlabel = \"%s\"\n"
	     "\t\trankdir = LR\n"
	     "\n",
	     k->id,
	     ((k->name && strlen(k->name))
	      ? isolat1_str(buffer_name, k->name, BUFFER_LENGTH) : k->id));

    graph_subnet(out, p, NULL, -1);

    fprintf( out, "\t} // end of subgraph '%s'\n\n", k->id);

    petrinet_all_out_flag(p, true);
  }

  return true;
}
/* End of graph_knot() */


flag_t
graph(FILE *out, petrinet_t *p)
{
  /*

  graph() - 

  */
  subnet_t *c;
  index_t i=0;

  fprintf( out, 
	   "//\n// Rootgraph\n//\n"
	   "subgraph cluster%i {\n"
	   "\tlabel = \"%s\"\n"
	   "\n",
	   i,
	   p->sn->id);

  i++;				/* first subnet '1' */

  for (c=&(p->sn[i]); c->id; c++) {	/* all subnets w/ id */
    if (c->valid) {
      graph_subnet(out, p, c, i);
      i++;
    }
  }

  write_knots(out, p, p->sn, state);

  write_knots(out, p, p->sn, transition);

  write_edges(out, p, p->sn);

  fprintf( out, "}\n\n");

  return true;
}
/* End of graph() */


void
graph_header(FILE *out, petrinet_t *p)
{
  /*

  graph_header() - 

  */

  fprintf( out, 
	   "digraph pkg2 {\n"
	   "//size=\"%i,%i\";\n"
	   "//overlap=\"false\";\n"
	   "//splines=\"true\";\n"
	   "//Damping=\"0.5\";\n"
	   "//ratio = \"fill\";\n"
	   "fontname = \"%s\"\n"
	   "fontsize = \"8\"\n"
	   "\n"
	   "node [\n"
	   "	fontname = \"%s\"\n"
	   "	fontsize = \"8\"\n"
	   "	fixedsize= \"true\"\n"
/* 	   "	shape = \"circle\"\n" */
	   "];\n"
	   "\n",
	   p->n_s, p->n_t, font_default, font_default
	   );
}
/* End of graph_header() */


void
graph_footer(FILE *out)
{
  /*

  graph_footer() - 

  */

  fprintf( out, "\n}\n");
}
/* End of graph_footer() */


