/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>
#include <limits.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <cgi-util.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"
#include "graph.h"
#include "web.h"
#include "pkg2.h"
#include "web_pkg2.h"


/* file name of the executable, used for <form> calls */
/* char *executable;		 */

int
main(int argc, char **argv) {

/*!
 */

/*   petrinet_t *pn; */
  pkg2_t *pkg2 = NULL;

  index_t i;
  knot_t *k;
  requirement_t *r;

  xmlDoc *xml;
  xmlNode *root_element;

  subnet_t *s;

/*   char *filename_petrinet; */
  char *filename_cfg;
/*   char *doc_prefix; */
  char *mode_str;
/*   char *flag_str; */

  run_t run_mode;

  int res;
  int n_paths;

  FILE *filep_subnet_graph;
  FILE *filep_dot;
  FILE *filep_txt;
  FILE *filep_navigation;
  FILE *filep_knot_table;

  char filename_dot[BUFFER_LENGTH];
  char filename_txt[BUFFER_LENGTH];
  char filename_knot_table[BUFFER_LENGTH];
  char filename_navigation[BUFFER_LENGTH];

  /****************************************************

   Options

  */

  /* run in command line mode if there is any argument in argv[]
   */

/*   run_mode = command_line; */
  run_mode = (argc > 1) ? command_line : cgi;

  if (run_mode == cgi) {	/* cgi run mode */

    /* Initialize the CGI */
    res = cgi_init();

    /* Was there an error initializing the CGI??? */
    printf_exit((res != CGIERR_NONE), 1, "%s", res, cgi_strerror(res));
  
    printf_debug("Run '%s' in CGI mode", argv[0]);

#if 0
    /* Grab some fields from an HTML form and display them:
     */
    /*! \todo evaluate argv[0] for executable */
    executable = cgi_getentrystr("cgi_call");
    printf_exit((executable == NULL),1,"No CGI call possible (define \"cgi_call\" in page)");
    printf_debug("CGI executable = '%s'", executable);

    doc_prefix = cgi_getentrystr("doc_prefix");
    printf_debug("doc_prefix = '%s'", pkg2->pn->doc_prefix);

    filename_petrinet = cgi_getentrystr("xml_name");
#endif
    filename_cfg = cgi_getentrystr("cfg_file");

  }
  else {
    printf_debug("Run in command line mode");

/*     executable = argv[0]; */

    filename_cfg = argv[1];
/*     filename_petrinet = "../../net/zf-cad/zf-cad.xml"; */

/*     filename_navigation = "navigation.html"; */
  }

  printf_exit((filename_cfg == NULL), 1, "No valid call");

  printf_debug("Read Configuration from '%s'", filename_cfg);

  /****************************************************

   !!! Construction area !!!

  */

  LIBXML_TEST_VERSION

  mem_init();			/* initialize my memory manager */

  pkg2 = pkg2_new(filename_cfg,run_mode);

  /****************************************************

   Processing

  */

  /****************************************************

   Output

  */

  if (run_mode == cgi) {

    mode_str = cgi_getentrystr("mode");
    printf_debug("mode = '%s'", mode_str);

    if (xmlStrEqual(mode_str,"Graph")) {
      /* output of graphiz code */
      printf("Content-type: text/%s\n\n", (1 ? "dot" : "plain"));
    }
    else {
      printf("Content-type: text/%s\n\n", (1 ? "html" : "plain"));
      web_header(stdout, pkg2->pn);
      web_form_mode(stdout, pkg2);
    }

    if (xmlStrEqual(mode_str,"Modelle")) {
      web_knots(stdout, pkg2->pn, state);
    }
    else if (xmlStrEqual(mode_str,"Prozesse")) {
      web_knots(stdout, pkg2->pn, transition);
    }
    else if (xmlStrEqual(mode_str,"Netze")) {
      web_subnets_list(stdout, pkg2->pn);
    }
    else if (xmlStrEqual(mode_str,"Anforderungen")) {
      web_requirements_table(stdout, pkg2);
    }
    else if (xmlStrEqual(mode_str,"Zuordnung")) {
      web_reqs_table(stdout, pkg2);
    }
    else if (xmlStrEqual(mode_str,"Hilfe")) {
    }
    else if (xmlStrEqual(mode_str,"Suche")) {
      /* new search */
      web_search(stdout, pkg2->pn);
    }
    else if (petrinet_set_start(pkg2->pn, (xmlChar *) cgi_getentrystr("start"))) {
      /* a start state found -- show the result of the search */
      if (petrinet_set_target(pkg2->pn, (xmlChar *) cgi_getentrystr("target"))) {
	/* target knot */

	petrinet_all_out_flag(pkg2->pn,false);

	/* mark all excluded subnets as not valid
	 */
	for (i = 0; i < cgi_getnumentries("subnets"); i++) {
	  char *sn_name = cgi_getnentrystr("subnets", i);
	  index_t sn_i = subnet_index4id(pkg2->pn,sn_name);
	  
	  printf_debug("exclude '%s' -> '%i'", sn_name, sn_i);
	  petrinet_subnet_valid(pkg2->pn, &(pkg2->pn->sn[sn_i]), false); 
	}

	/* mark all excluded transitions
	 */
	for (i = 0; i < cgi_getnumentries("transition"); i++) {
	  char *t_name = cgi_getnentrystr("transition", i);
	  index_t t_i = transition_index4id(pkg2->pn,t_name);

	  printf_debug("exclude '%s' -> '%i'", t_name, t_i);
	  pkg2->pn->t[t_i].valid = false;
	}

	/* mark all excluded states
	 */
	for (i = 0; i < cgi_getnumentries("state"); i++) {
	  char *s_name = cgi_getnentrystr("state", i);
	  index_t s_i = state_index4id(pkg2->pn, s_name);

	  printf_debug("exclude '%s' -> '%i'", s_name, s_i);
	  pkg2->pn->s[s_i].valid = false;
	}

	if (xmlStrEqual(mode_str,"Start")) {
	  /* output of path table */
	  web_pathnet_req(stdout, pkg2);
	  /* 	web_pathnet(stdout, pkg2->pn, doc_prefix); */
	}
	else {
	  /* output of Graphiz code */
/* 	  fprintf(stdout,"<pre>\n\n"); */
	  graph_header(stdout, pkg2->pn);
	  graph_pathnet(stdout, pkg2->pn);
	  graph_footer(stdout);
/* 	  fprintf(stdout,"</pre>\n"); */
	}
      }
    }
    else {
      /* default for cgi mode */
      web_subnets_list(stdout, pkg2->pn);
    }

    if (xmlStrEqual(mode_str,"Graph")) {
    }
    else {
      web_footer(stdout, pkg2->pn);
    }
  }
  else {

    if (mode_images(pkg2)) {
      printf_debug("Whole graph '%s'", pkg2->pn->id);

      sprintf(filename_dot, "%s%s.dot", pkg2->pn->doc_prefix, (char *) pkg2->pn->id);

      filep_dot = fopen(filename_dot, "w");

      printf_exit((! filep_dot), 2, "Cant create '%s'", filename_dot );

      graph_header(filep_dot, pkg2->pn);

      graph(filep_dot, pkg2->pn);

      graph_footer(filep_dot);

      fclose(filep_dot);

      graph_dot_call(pkg2->pn->id, pkg2->pn->doc_prefix);

      web_subnet_card_file(pkg2->pn, pkg2->pn->sn);

      petrinet_all_out_flag(pkg2->pn,true);

      /* subnet cards */
      for (s=&(pkg2->pn->sn[1]); s->id; s++) {
	graph_subnet_file(pkg2->pn, s, 1);
	graph_dot_call(s->id, pkg2->pn->doc_prefix);
	web_subnet_card_file(pkg2->pn, s);
      }
    }

    if (mode_requirements(pkg2)) {
      sprintf(filename_navigation, "%s%s.html", pkg2->pn->doc_prefix, "requirements");

      filep_navigation = fopen(filename_navigation, "w");

      printf_exit( (! filep_navigation), 2, "Cant create '%s'", filename_navigation );

      web_header(filep_navigation, pkg2->pn);
      web_requirements_table(filep_navigation, pkg2);
      web_footer(filep_navigation, pkg2->pn); 

      sprintf(filename_navigation, "%s%s.html", pkg2->pn->doc_prefix, "reqs");

      filep_navigation = fopen(filename_navigation, "w");

      printf_exit( (! filep_navigation), 2, "Cant create '%s'", filename_navigation );

      web_header(filep_navigation, pkg2->pn);
      web_reqs_table(filep_navigation, pkg2);
      web_footer(filep_navigation, pkg2->pn); 
    }

#if 1
    sprintf(filename_navigation, "%s%s.html", pkg2->pn->doc_prefix, "navigation");

    filep_navigation = fopen(filename_navigation, "w");

    printf_exit( (! filep_navigation), 2, "Cant create '%s'", filename_navigation );

    web_header(filep_navigation, pkg2->pn);
    web_search(filep_navigation, pkg2->pn);
    web_footer(filep_navigation, pkg2->pn);

    fclose(filep_navigation);
#endif

#if 1
    sprintf(filename_knot_table,"%s%s-prozesse.html", pkg2->pn->doc_prefix, (char *) pkg2->pn->id);

    filep_knot_table = fopen(filename_knot_table, "w");

    printf_exit( (! filep_knot_table), 2, "Cant create '%s'", filename_knot_table );

    web_header(filep_knot_table, pkg2->pn);
/*     web_knots_table(filep_knot_table, pkg2->pn, empty_str); */
    web_knots(filep_knot_table, pkg2->pn, transition);
    web_footer(filep_knot_table, pkg2->pn);

    fclose(filep_knot_table);
#endif

#if 1
    sprintf(filename_knot_table,"%s%s-modelle.html", pkg2->pn->doc_prefix, (char *) pkg2->pn->id);

    filep_knot_table = fopen(filename_knot_table, "w");

    printf_exit( (! filep_knot_table), 2, "Cant create '%s'", filename_knot_table );

    web_header(filep_knot_table, pkg2->pn);
/*     web_knots_table(filep_knot_table, pkg2->pn, empty_str); */
    web_knots(filep_knot_table, pkg2->pn, state);
    web_footer(filep_knot_table, pkg2->pn);

    fclose(filep_knot_table);
#endif

#if 1
    for (k=pkg2->pn->s; k->id; k++) {
      if (mode_images(pkg2)) {
	graph_knot_file(pkg2->pn, k);
	graph_dot_call(k->id, pkg2->pn->doc_prefix);
      }
      web_knot_card_req_file(pkg2, k);
    }
    for (k=pkg2->pn->t; k->id; k++) {
      if (mode_images(pkg2)) {
	graph_knot_file(pkg2->pn, k);
	graph_dot_call(k->id, pkg2->pn->doc_prefix);
      }
      web_knot_card_req_file(pkg2, k);
    }
    for (r=pkg2->r; mode_requirements(pkg2) && r->id; r++) {
      web_requirements_card_file(pkg2, r);
    }
#endif
  }
  
  /****************************************************

   !!! Destruction area !!!

  */

  if (pkg2)
    pkg2_free(pkg2);

  xmlCleanupParser();

  if (run_mode == cgi)     /* Close up the CGI: */
    cgi_quit();

  exit(EXIT_SUCCESS);

  return EXIT_SUCCESS;		/* dummy only */
}

