/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/* 

 */

/****************************************************************************

  type definition: subnet_t
*/

/*! structure type for subnets
 */
typedef struct {

  xmlChar *id;			/*!< id of subnet */

  xmlChar *name;		/*!< name of subnet */

  xmlChar *description;		/*!< description of subnet */

  flag_t valid;			/*!< valid flag */

} subnet_t;


/****************************************************************************

  type definition: 
*/

typedef enum {state, transition} knot_type_t;

/* ***************************************************************************

  type definition: knot_t
 */

/*! knot structure
 */
typedef struct {
 
  knot_type_t type;             /*!< type of element */

  xmlChar *id;			/*!< id of subnet */

  xmlChar *name;		/*!< name of subnet */

  xmlChar *description;		/*!< description of subnet */

  flag_t valid;			/*!< valid flag */

  subnet_t *sn;			/*!< parent subnet */

/*   req_t *req; */

  xmlChar *color;

  xmlChar *rating;

  flag_t out;			/* flag for output */

  index_t i;                    /* redundant but performant */

  htitem_t *l;			/* array of links */

} knot_t;


/* ***************************************************************************

   preprocessor macros for access to knot_t
 */

#define statep(N) (N != NULL && N->type == state && N->valid)

#define transitionp(N) (N != NULL && N->type == transition && N->valid)

/* warning: not nice -> better edgep() */

#define edge_st(P,s,t) (P->e[s][t] & 1)

#define edge_ts(P,t,s) (P->e[s][t] & 2)

#define knotid(N) N->id

#define knotname(N) N->name

#define state_index(N) (statep(N) ? N->i : ERROR_INDEX)

#define transition_index(N) (transitionp(N) ? N->i : ERROR_INDEX)

/* ***************************************************************************

  type definition: edge_t
 */

typedef struct {
 
  knot_t *from;
  knot_t *to;

  xmlChar *description;

  xmlChar *color;

  xmlChar *style;

/*   req_t *req; */

  flag_t valid;			/* flag for validity of edge */

  flag_t out;			/* flag for output */

  subnet_t *sn;			/* parent subnet */

  htitem_t *l;			/* array of links */

} edge_t;


/****************************************************************************

  type definition: petrinet_t
*/

typedef struct {

  /* 

  */

  xmlDoc *xml;

  xmlNode *root_element;

  xmlNode *cfg_root_element;

  xmlChar *id;

  xmlChar *version;

  xmlChar *name;

  xmlChar *description;

  xmlChar *copyright;

  xmlChar *remark;

  xmlChar *contact;

  index_t n_sn;
  subnet_t *sn;			/* subnets */

  index_t n_s;
  knot_t *s;

  index_t n_t;
  knot_t *t;

  index_t n_e;			/* number of edges */
  edge_t *edge;			/* array of edges */
  index_t **e;                  /* 2D matrix of edges */

  knot_t **path;                        /* pointer for path buffer */
  index_t n_paths; /* number of found path between 'start' and 'target' */
  index_t l_path;               /* last knot index in path */

  knot_t *start;

  knot_t *target;

  char doc_prefix[BUFFER_LENGTH];
  char *html_prefix;
  char *executable;
  char *filename;		/* name of parsed cfg xml file -> CGI */
  char *filename_absolut;		/* name of parsed cfg xml file -> CGI */

  run_t mode_run;		/*! run mode for petrinet */

} petrinet_t;


/****************************************************************************

   preprocessor macros for access to petrinet_t
*/


/****************************************************************************

   Declaration of public constants
*/

extern const xmlChar *tag_state;

extern const xmlChar *tag_transition;

extern const xmlChar *tag_edge;

extern const xmlChar *tag_pnet;

extern const xmlChar *tag_subnet;

extern const xmlChar *tag_hitem;


/****************************************************************************

   preprocessor macros for access to petrinet_t
*/

#define petrinet_set_start(P,ID) set_start_target(P,0,ID)

#define petrinet_set_target(P,ID) set_start_target(P,1,ID)

#define mode_cgi(P) (P->mode_run == cgi)


/****************************************************************************

   Declaration of public functions (prototypes)
*/

extern petrinet_t *
petrinet_new(char *filename, run_t run_mode);

extern flag_t 
set_start_target(petrinet_t *p, int end, xmlChar *start_id);

extern flag_t 
path_forward(petrinet_t *p);

extern flag_t 
branch_backward(petrinet_t *p);

extern int 
path_next(void);

extern index_t 
state_index4id(petrinet_t *p, xmlChar *id);

extern index_t 
transition_index4id(petrinet_t *p, xmlChar *id);

extern index_t
subnet_index4id(petrinet_t *p, xmlChar *id);

extern index_t
edge_index4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to);

extern void
petrinet_free(petrinet_t *p);

extern flag_t
path_exist(petrinet_t *p, xmlChar *start_id, xmlChar *target_id);

extern flag_t
petrinet_center_valid(petrinet_t *p, xmlChar *s_id);

extern flag_t
petrinet_pathnet_out_flag(petrinet_t *p, flag_t f);

extern void
petrinet_all_out_flag(petrinet_t *p, flag_t f);

extern void
petrinet_neighbors_out_flag(petrinet_t *p, knot_t *k, flag_t f);

extern knot_t *
state_ptr4id(petrinet_t *p, xmlChar *id);

extern knot_t *
transition_ptr4id(petrinet_t *p, xmlChar *id);

extern edge_t *
edge_ptr4id(petrinet_t *p, xmlChar *id_from, xmlChar *id_to);

