/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/****************************************************************************

  type definition: requirement_t
*/

/*! description of a special requirement
 */
typedef struct {

  xmlChar *id;			/*!< id of subnet */

  xmlChar *name;		/*!< name of subnet */

  xmlChar *description;		/*!< description of subnet */

  subnet_t *sn;			/*!< parent subnet */

  htitem_t *l;			/*!<  array of links */

  htitem_t *check;		/*!<  array of checks */

  flag_t valid;			/*!< valid flag */

  flag_t out;			/*!<  flag for output */

} requirement_t;


/****************************************************************************

  type definition: req_t
*/

typedef struct {

  /* reference from an edge to of a special requirement

  */

  xmlChar *ref;			/*!< id reference to a requirement_t */

  xmlChar *cat;			/*!< category */

  edge_t *edge;			/*!< source reference edge */

  knot_t *knot;			/*!< or knot */

  flag_t valid;			/*!< valid flag */

} req_t;


/****************************************************************************

  type definition: pkg2_t
*/

typedef struct {

  /* 

  */

  petrinet_t *pn;

  index_t n_r;
  requirement_t *r;

  index_t n_req;
  req_t *req;

  int mode;			/*!< bit field for modules flags */

/*   char *filename_cfg; */

} pkg2_t;



extern pkg2_t *
pkg2_new(char *filename, run_t run_mode);

extern void
pkg2_free(pkg2_t *pkg2);

extern void
pkg2_all_out_flag(pkg2_t *pkg2, flag_t f);

extern void
pkg2_knot_requirement_out_flag(pkg2_t *pkg2, knot_t *k, flag_t f);

extern char 
cat4edge(pkg2_t *pkg2, knot_t *k_a, knot_t *k_b, requirement_t *r);

extern char 
cat4knot(pkg2_t *pkg2, knot_t *k_a, requirement_t *r);

/* no 'bit shifting' because portability */
#define FLAG_CGI          1
#define FLAG_DEVELOP      2 * FLAG_CGI
#define FLAG_REQUIREMENTS 2 * FLAG_DEVELOP
#define FLAG_LINKS        2 * FLAG_REQUIREMENTS
#define FLAG_DOCS         2 * FLAG_LINKS
#define FLAG_IMAGES       2 * FLAG_DOCS

#define mode_develop(PKG2)      (PKG2->mode & FLAG_DEVELOP)
#define mode_links(PKG2)        (PKG2->mode & FLAG_LINKS)
#define mode_requirements(PKG2) (PKG2->mode & FLAG_REQUIREMENTS)
#define mode_docs(PKG2)         (PKG2->mode & FLAG_DOCS)
#define mode_images(PKG2)       (PKG2->mode & FLAG_IMAGES)

