/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/



#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>
#include <limits.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"
#include "version.h"

#include "web.h"

/****************************************************************************

   Declaration of private functions (prototypes)
*/

static flag_t
write_subnets_form(FILE *out, petrinet_t *p);

/****************************************************************************

 Definition of functions
*/


flag_t
write_subnets_form(FILE *out, petrinet_t *p)
{
  /*

  write_subnets_form() - 

  */
  subnet_t *c;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\">\n<tbody>\n");
  
  fprintf( out, "<tr><th colspan=\"1\">%s</th></tr>\n", "Subnets");

  for (c=&(p->sn[1]); c->id; c++) {	/* all subnets w/ id */
    if (c->valid) {

      fprintf( out, "<tr><td><input type=\"checkbox\" name=\"subnets\" value=\"%s\">",
	       c->id);

      fprintf( out, "%s</td></tr>\n", form_subnet(buffer_name, p, c));

/*       fprintf( out, "<a href=\"%s%s%s.html\">%s</a></td></tr>\n", */
/* 	       doc_prefix, */
/* 	       DIR_DOC, */
/* 	       c->id, */
/* 	       isolat1_str(buffer_name, c->name, BUFFER_LENGTH)); */
    }
  }

  fprintf( out, "</tbody></table>\n");

  return true;
}
/* End of write_subnets_form() */



flag_t
write_knots_form(FILE *out, petrinet_t *p, knot_type_t type)
{
  /*

  write_knots_form() - 

  */
  knot_t *c;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\"><tbody>\n");
  
  fprintf( out, "<tr><th colspan=\"1\">%s</th></tr>\n",
	   ((type == state) ? "States" : "Transitions"));

  for (c=((type == state) ? p->s : p->t); c->id; c++) /* all knots w/ id */
    if (c->valid && c->sn->valid && c->out)
      fprintf( out, "<tr><td><input type=\"checkbox\" name=\"%s\" value=\"%s\">%s</td></tr>\n",
	       ((c->type == state) ? "state" : "transition"),
	       c->id,
	       form_knot(buffer_name, p, c));

  fprintf( out, "</tbody></table>\n");

  return true;
}
/* End of write_knots_form() */


/* list of htitem_t */
flag_t
web_htitem_list(FILE *out, htitem_t *l)
{
  unsigned char buffer_name[BUFFER_LENGTH];
  int j;

  if (l) {
    fprintf( out, "<ul>\n");

    for (j=0; l[j].text; j++) {
      fprintf( out, "<li>");
      if (l[j].href)
	fprintf( out, "<a target=\"_blank\" href=\"%s\">", (char *) l[j].href);

      fprintf( out, "%s", isolat1_str(buffer_name, l[j].text, BUFFER_LENGTH) );

      if (l[j].href)
	fprintf( out, "</a>");

      fprintf( out, "</li>");
    }
    fprintf( out, "</ul>\n");
  }

  return true;
}
/* End of web_htitem_list() */


/*! writes a table with all knots of 'type'

\param out Output file handle
\param p pointer to used petrinet
\param type of knots in table
\param doc_prefix for link

\result flag only
*/
flag_t
web_knots(FILE *out, petrinet_t *p, knot_type_t type)
{
  /* write_knots() - 
  */

  knot_t *c;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\"><tbody>\n");
  
  fprintf( out, "<tr><th colspan=\"3\">%s</th></tr>\n", ((type == state) ? "States" : "Transitions"));

  for (c=((type == state) ? p->s : p->t); c->id; c++) { /* all knots w/ id */
    if (c->valid && c->sn->valid && c->out) {
      fprintf( out, "<tr>");

#if 0
      fprintf( out, "<td><a href=\"%s%s.html\">%s</a></td>\n",
             p->doc_prefix,
             c->id,
             isolat1_str(buffer_name, c->name, BUFFER_LENGTH));
#else
      fprintf( out, "<td>%s</td>\n", form_knot(buffer_name,p,c));
#endif

      fprintf( out,
	       "<td>"
	       "<a href=\"%s%s.html\">%s</a>"
	       "</td>\n\n",
	       p->doc_prefix,
	       c->sn->id,
	       ((c->sn->name && strlen((char *) c->sn->name)) ? isolat1_str(buffer_name, c->sn->name, BUFFER_LENGTH) : c->sn->id)
	       );

      fprintf( out, "<td>%s</td>",
	       isolat1_str(buffer_name, c->description, BUFFER_LENGTH));

      fprintf( out, "</tr>\n");
    }
  }

  fprintf( out, "</tbody><table>\n");

  return true;
}
/* End of write_knots() */


/*! writes a html table of all path's

\param out Output file handle
\param p pointer to used petrinet
\param doc_prefix for link

\result flag only
 */
flag_t
web_pathnet(FILE *out, petrinet_t *p)
{
  /* 
  */
  int i = 0;

  if (p->start && p->start->valid
      && p->target && p->target->valid) {

    do {
      if (path_forward(p)) {
	i++;
	if (i< 10) {
	  web_path(out, p);
	}
      }
    } while (branch_backward(p));

    printf_debug("%i path between '%s' and '%s' found",
	    p->n_paths, p->start->id, p->target->id);
  }

  return (p->n_paths > 0);
}
/* End of web_pathnet() */


/*! writes a html table of the current path

\param out Output file handle
\param p pointer to used petrinet
\param doc_prefix for link

\result flag only
*/
flag_t
web_path(FILE *out, petrinet_t *p)
{
  int i, j, m;
  knot_t *k;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\"><tbody>\n");

  fprintf( out, "<tr><th colspan=\"5\">%s</th></tr>\n", "Path");
  
  for (i=0, m=0; i <= p->l_path; i++) {

    k = p->path[i];

    fprintf( out, "<tr>\n");

    if (k->type == state) {
      fprintf( out,
	       "<td></td>"
	       "<td class=\"%s\">%s</td>\n",
	       "state",
	       form_knot(buffer_name, p, k));
    }
    else {
      m++;
      fprintf( out,
	       "<td>%i</td>"
	       "<td class=\"%s\">%s</td>\n",
	       m,
	       "transition",
	       form_knot(buffer_name, p, k));
    }

    fprintf( out,
	     "<td>%s</td>\n",
	     form_subnet(buffer_name, p, k->sn));

    fprintf( out, "<td>%s</td>\n",
	     isolat1_str(buffer_name, k->description, BUFFER_LENGTH));
  
    /* link list */
    fprintf( out, "<td>\n");
    web_htitem_list(out, k->l);
    fprintf( out, "</td>\n");

    fprintf( out, "</tr>\n");
  }
  fprintf( out, "</tbody><table>\n");

  return true;
}
/* End of web_path() */


/*! web_subnet_card_file() - 

*/
flag_t
web_subnet_card_file(petrinet_t *p, subnet_t *s)
{

  assert(s);

  if (s->id && s->valid) {

    char card_name[BUFFER_LENGTH];
    FILE *filep_subnet_card;

    assert(s->name);

    sprintf(card_name, "%s%s.html", p->doc_prefix, (char *) s->id);

    filep_subnet_card = fopen(card_name, "w");

    printf_exit( (! filep_subnet_card), 2, "Cant create '%s'", card_name );

    printf_debug("Write '%s'", card_name);

    web_header(filep_subnet_card, p);

    /*! \todo insert form_mode() */

    fprintf(filep_subnet_card,
	    "<h3>Subnet: %s</h3>\n",
	    form_subnet(card_name, p, s));

    /*     if (mode_imgs(pkg2)) { */
    /* Graph with sensitive png */
    fprintf(filep_subnet_card, "<center>");
    web_card_img(filep_subnet_card, s->id, p->doc_prefix);
    fprintf(filep_subnet_card, "</center>");
    /*     } */

    web_footer(filep_subnet_card, p);

    fclose(filep_subnet_card);

    return true;
  }

  return false;
} /* End of web_subnet_card_req_file() */


/*! web_knot_card_file() - 
 */
flag_t
web_knot_card_file(petrinet_t *p, knot_t *k)
{
  assert(k);

  if (k->id && k->valid && k->sn->valid) {

    char card_name[100];
    FILE *filep_knot_card;

    assert(k->name);

    sprintf(card_name, "%s%s.html", p->doc_prefix, (char *) k->id);

    filep_knot_card = fopen(card_name, "w");

    printf_exit( (! filep_knot_card), 2, "Cant create '%s'", card_name );

    printf_debug("Write '%s'", card_name);

    web_header(filep_knot_card, p);

    web_knot_card(filep_knot_card, p, k);

    web_footer(filep_knot_card, p);

    fclose(filep_knot_card);

    return true;
  }

  return false;
} /* End of web_knot_card_file() */


/*! writes a sensitive image section for the simple cardfile for the id

\param out output file handle
\param id to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
*/
flag_t
web_card_img(FILE *out, xmlChar *id, char *doc_prefix)
{
  index_t i;
  unsigned char buffer_name[BUFFER_LENGTH];
  FILE *filehandle_map;

  if (id && strlen((char *) id)) {

    sprintf(buffer_name, "%sgraph_%s.map", doc_prefix, (char *) id);

    if ((filehandle_map = fopen(buffer_name,"r"))) {
      printf_debug("Read from Map File '%s'", buffer_name);

      fprintf( out, "<map name=\"%s\">\n", id);

      while (fgets(buffer_name, BUFFER_LENGTH-1, filehandle_map)) {
	fputs(buffer_name,out);	
      }

      fprintf( out,
	       "</map>\n"
	       "<img usemap=\"#%s\" src=\"graph_%s.png\">\n",
	       id,
	       id);

      fclose(filehandle_map);
    }
    else {
      printf_debug("Map-File '%s' not found", buffer_name);

      fprintf( out,
	       "<img src=\"graph_%s.png\">\n",
	       id);
    }

#if 0
    fprintf( out,
	     "<a target=\"_blank\" href=\"%s.pdf\">PDF</a>\n",
	     id);
#endif
  }

  return false;
}
/* End of web_card_img() */


/*! writes a sensitive image section for the simple cardfile for the id

\param out output file handle
\param id to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
*/
flag_t
web_card_doc(FILE *out, xmlChar *id, char *doc_prefix)
{
  flag_t flag_body = false;
  unsigned char buffer[BUFFER_LENGTH];
  FILE *filehandle;

  if (id && strlen((char *) id)) {

    sprintf(buffer, "%s%s.html", doc_prefix, (char *) id);

    if ((filehandle = fopen(buffer,"r"))) {
      printf_debug("Read from '%s'", buffer);

/*       fprintf( out, "<h3>Dokumentation</h3>\n"); */

      while (fgets(buffer, BUFFER_LENGTH-1, filehandle)) {

	/*! \todo pattern for body attributes and remarks in html code */
	if (strstr(buffer,"<body")) {
	  flag_body = true;
	}
	else if (strstr(buffer,"</body>")) {
	  flag_body = false;
	}
	else if (flag_body) {
	  fputs(buffer,out);	
	}
      } 
      fclose(filehandle);
    }
    else {
      printf_debug("File '%s' not found", buffer);

      /* try to find an image file for id */
      sprintf(buffer, "%s%s.png", doc_prefix, (char *) id);

      if ((filehandle = fopen(buffer,"r"))) {
	printf_debug("Found '%s'", buffer);

/* 	fprintf( out, "<img src=\"%s\">\n", buffer); */
	fprintf( out, "<img src=\"%s.png\">\n", (char *) id);

	fclose(filehandle);
      }
      else {
	printf_debug("File '%s' not found", buffer);
      }

      /* try to find a text file (documentation draft) for id */
      sprintf(buffer, "%s%s.txt", doc_prefix, (char *) id);

      if ((filehandle = fopen(buffer,"r"))) {
	printf_debug("Found '%s'", buffer);

	fprintf( out, "<pre>\n");

	while (fgets(buffer, BUFFER_LENGTH-1, filehandle)) {
	  fputs(buffer,out);	
	}
	fclose(filehandle);

	fprintf( out, "</pre>\n");
      }
      else {
	printf_debug("File '%s' not found", buffer);
      }
    }
  }

  return false;
}
/* End of web_card_doc() */


/*! writes a knot card of the knot 'k'

\param out Output file handle
\param p pointer to used petrinet
\param k knot
\param doc_prefix for link

\result flag only
*/
flag_t
web_knot_card(FILE *out, petrinet_t *p, knot_t *k)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  if (k->id && k->valid && k->sn->valid) {

    fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\" width=\"80%%\"><tbody>\n");
  
    fprintf( out,
	     "<tr><th>%s \n",
	     ((k->name && strlen((char *) k->name))
	      ? isolat1_str(buffer_name, k->name, BUFFER_LENGTH) : k->id));

    fprintf( out,
	     "(%s)</th></tr>\n",
	     ((k->sn->name && strlen((char *) k->sn->name)) ? isolat1_str(buffer_name, k->sn->name, BUFFER_LENGTH) : k->sn->id) );

    fprintf( out, "<tr><td>%s</td></tr>\n",
             isolat1_str(buffer_name, k->description, BUFFER_LENGTH));

    fprintf( out, "<tr><td><img src=\"graph_%s.png\"></td></tr>\n",
             k->id);

    fprintf( out, "<tr><td>Links\n");
    web_htitem_list(out, k->l);
    fprintf( out, "</td></tr>\n");

    fprintf( out, "</tbody><table>\n");
  }

  return true;
}
/* End of web_knot_card() */


void
web_search(FILE *out, petrinet_t *p)
{
  /*

  write_navigation() - 

  */
  knot_t *k; 
  char *selects[] = {"start","target", NULL};
  index_t j;
  unsigned char buffer_name[BUFFER_LENGTH];

#if 0
  fprintf( out, 
	   "<form action=\"%s\" method=\"post\">\n",
	   p->executable
	   );
#endif

  fprintf( out, 
	   "<form action=\"%s\" method=\"post\">\n"
	   "   <input type=\"hidden\" name=\"cfg_file\" value=\"%s%s\">\n",
	   p->executable,
	   p->filename_absolut, p->filename
	   );

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"5\"><tbody>\n");

  fprintf( out, "<tr><td colspan=\"3\">\n");

#if 0
  /*! \bug input on xml configuration */
  fprintf( out, 
	   "   <input type=\"hidden\" name=\"cgi_call\" value=\"%s\">\n"
	   "   <input type=\"hidden\" name=\"doc_prefix\" value=\"%s\">\n"
	   "   <input type=\"hidden\" name=\"xml_name\" value=\"%s\">\n",
	   p->executable,
	   p->doc_prefix,
	   p->filename
	   );
#endif

  /* two select menues */
  for (j=0; selects[j]; j++) {
    fprintf( out, 
	     "%s<select name=\"%s\" size=\"1\">\n",
	     selects[j], selects[j]
	     );

    for (k=p->s; k->id; k++) {
      if (k->valid) {
	fprintf( out, 
		 "<option value=\"%s\">%s / %s\n",
		 k->id, 
		 k->sn->id,
		 ((k->name && strlen((char *) k->name)) ? isolat1_str(buffer_name, k->name, BUFFER_LENGTH) : k->id)
		 );
      }
    }

    fprintf( out, "</select>");
  }

  fprintf( out, "<input type=\"submit\" name=\"mode\" value=\"Start\">\n");

  fprintf( out, "<input type=\"submit\" name=\"mode\" value=\"Graph\">\n");

  fprintf( out, "</td></tr>\n");


  fprintf( out, "<tr><td valign=\"top\">\n");

  write_subnets_form(out, p);

  fprintf( out, "</td><td valign=\"top\">\n");

  write_knots_form(out, p, state);

  fprintf( out, "</td><td valign=\"top\">\n");

  write_knots_form(out, p, transition);

  fprintf( out, "</td></tr>\n");

  fprintf( out, "</tbody></table>\n");

  fprintf( out, 
	   "</form>\n"
	   );

#if 0
  fprintf( out, 
    	   "<hr>\n"
    	   "<p>XML:</p>\n"
    	   "<ul>\n"
      	   "<li><a href=\"\">Download</a></li>\n"
      	   "<li><a href=\"\">Download</a> (valid)</li>\n"
      	   "<li><a href=\"\">Upload</a></li>\n"
      	   "<li><a href=\"\">Aktualisieren</a></li>\n"
      	   "<li></li>\n"
      	   "<li><a href=\"\">Download</a> aktueller Zustand</li>\n"
      	   "<li><a href=\"\">Download</a> DOT-File</li>\n"
      	   "<li><a href=\"\">Download</a> Graph</li>\n"
    	   "</ul>\n"
	   );

  fprintf( out, 
    	   "<p>DOT:</p>\n"
    	   "<ul>\n"
      	   "<li><a href=\"\">Download</a></li>\n"
      	   "<li><a href=\"\">Upload</a></li>\n"
      	   "<li><a href=\"\">Aktualisieren</a></li>\n"
      	   "<li></li>\n"
      	   "<li><a href=\"\">Download</a> aktueller Zustand</li>\n"
      	   "<li><a href=\"\">Download</a> DOT-File</li>\n"
      	   "<li><a href=\"\">Download</a> Graph</li>\n"
    	   "</ul>\n"
	   );
#endif

}
/* End of web_search() */


void
web_subnets_list(FILE *out, petrinet_t *p)
{
  subnet_t *s = p->sn;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, 
 	   "<p>Net: <a href=\"%s%s.html\">%s</a></p>",
	   p->doc_prefix,
/* 	   DIR_DOC, */
	   s->id,
	   ((s->name && strlen((char *) s->name)) ? isolat1_str(buffer_name, s->name, BUFFER_LENGTH) : s->id)
	   );

  fprintf( out, 
 	   "<p>Subnets<ul>\n");

  /* subnets */
  for ( s++; s->id; s++) {

    if (s->valid) {
      /*
      */
      fprintf( out,
	       "<li>"
	       "<a href=\"%s%s.html\">%s</a>"
	       "</li>\n\n",
	       p->doc_prefix,
/* 	       DIR_DOC, */
	       s->id,
	       ((s->name && strlen((char *) s->name)) ? isolat1_str(buffer_name, s->name, BUFFER_LENGTH) : s->id)
	       );
    }
  }

  fprintf( out,
	   "</ul></p>\n");

}
/* End of write_subnets_list() */


/*! forms a string with html anchor to the given buffer
 */
unsigned char *
form_knot(unsigned char *buffer, petrinet_t *p, knot_t *k)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  assert(p);
  assert(k);

  if (k->id) {
    if (p->doc_prefix) {
      sprintf( buffer, 
	       "<a href=\"%s%s.html\">%s</a>",
	       ((p->mode_run == cgi) ? p->doc_prefix : empty_str),
	       k->id,
	       isolat1_str(buffer_name, ((k->name && xmlStrlen(k->name) > 2) ? k->name : k->id), BUFFER_LENGTH)
	       );
    }
    else {
      /*! \todo non href marker */
      sprintf( buffer, 
	       "<a>%s</a>",
	       isolat1_str(buffer_name, ((k->name && xmlStrlen(k->name) > 2) ? k->name : k->id), BUFFER_LENGTH)
	       );
    }
  }

  return buffer;
}
/* Ende von form_knot() */


/*! forms a string with html anchor to the given buffer
 */
unsigned char *
form_subnet(unsigned char *buffer, petrinet_t *p, subnet_t *s)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  assert(p);
  assert(s);

  if (s->id) {
    if (p->doc_prefix) {
      sprintf( buffer, "<a href=\"%s%s.html\">%s</a>",
	       ((p->mode_run == cgi) ? p->doc_prefix : empty_str),
	       s->id,
	       isolat1_str(buffer_name, ((s->name && xmlStrlen(s->name) > 2) ? s->name : s->id), BUFFER_LENGTH)
	       );
    }
    else {
      /*! \todo non href marker */
      sprintf( buffer, "<a>%s</a>",
	       isolat1_str(buffer_name, ((s->name && xmlStrlen(s->name) > 2) ? s->name : s->id), BUFFER_LENGTH)
	       );
    }
  }

  return buffer;
}
/* Ende von form_subnet() */


void
web_header(FILE *out, petrinet_t *p)
{
  /*

  write_header() - 

  */

  unsigned char buffer_name[BUFFER_LENGTH];

  assert(p);

  fprintf( out, 
	   "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
           "\n"
	   "<html>\n"
	   "  <head>\n"
	   "    <link rel=\"stylesheet\" media=\"screen, print\" type=\"text/css\" href=\"%s%s\">\n"
	   "    <title>%s</title>\n"
	   "  </head>\n"
	   "  <body>\n"
	   "\n",
	   ((p->mode_run == cgi) ? p->doc_prefix : empty_str), "pkg2.css",
	   isolat1_str(buffer_name, ((p->name && xmlStrlen(p->name) > 2) ? p->name : p->id), BUFFER_LENGTH));
}
/* End of write_header() */


void
web_footer(FILE *out, petrinet_t *p)
{
  /*

  write_footer() - 

  */

  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out,
	   "<hr>"
	   "<p><tt><a href=\"%s\">%s</a></tt></p>\n"
	   "<p>Petrinet '%s' Version %s:<br> Copyright (C) %s,<br> Contact %s,<br> %s</p>\n"
 	   "</body>\n"
	   "</html>\n",
	   version_url, version_str,
	   p->id, p->version, p->copyright, p->contact, isolat1_str(buffer_name, p->remark, BUFFER_LENGTH)
	   );
}
/* End of write_footer() */


#if 0

flag_t
write_edges(petrinet_t *p)
{
  /* write_edges() - 

  */

  index_t s_i;
  index_t t_i;
  
  printf("//\n// Edges\n//\n");

  for (s_i=0; s_i < p->n_s; s_i++) {
    for (t_i=0; t_i < p->n_t; t_i++) {
      if (edge_st(p,s_i,t_i)) {
        printf("\t\"%s\" -> \"%s\";\n", p->s[s_i].id, p->t[t_i].id);
      }
      if (edge_ts(p,t_i,s_i)) {
        printf("\t\"%s\" -> \"%s\";\n", p->t[t_i].id, p->s[s_i].id);
      }
    }
  }

  return true;
}
/* End of write_edges() */


flag_t
write_partners(petrinet_t *p, knot_t *k)
{
  /* write_partners() - 

  */

  if (statep(k)) {
    index_t t_i;
    knot_t *t = p->t;

    /* all edges TO state k */
    printf("From: ");
    for (t_i=0; t_i < p->n_t; t_i++) {
      /* if (edge_ts(p,transition_index(t),state_index(k))) { */
      if (edgep(p,t,k)) {
        printf("%s ", knotname(t));
      }
      t++;
    }

    /* all edges FROM state k */
    printf(" | To: ");
    t = p->t;
    for (t_i=0; t_i < p->n_t; t_i++) {
      /*     if (edge_st(p,state_index(k),transition_index(t))) { */
      if (edgep(p,k,t)) {
        printf("%s ", knotname(t));
      }
      t++;
    }
  }
  else if (transitionp(k)) {
    index_t s_i;
    knot_t *s = p->s;

    /* all edges TO transition k */
    printf("From: ");
    for (s_i=0; s_i < p->n_s; s_i++) {
      /* if (edge_st(p,state_index(s),transition_index(k))) { */
      if (edgep(p,s,k)) {
        printf("%s ", knotname(s));
      }
      s++;
    }

    /* all edges FROM transition k */
    printf(" | To: ");
    s = p->s;
    for (s_i=0; s_i < p->n_s; s_i++) {
      /*     if (edge_ts(p,transition_index(k),state_index(s))) { */
      if (edgep(p,k,s)) {
        printf("%s ", knotname(s));
      }
      s++;
    }
  }
  else {
    /* error */
  }

  return true;
}
/* End of write_partners() */

#endif
