/*

  Copyright (C) 1999-2005 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>
#include <limits.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <cgi-util.h>

#include "tools.h"
#include "mem.h"
#include "petrinet.h"
#include "graph.h"
#include "web.h"
#include "pkg2.h"
#include "web_pkg2.h"

static unsigned char *
form_requirement(unsigned char *buffer, pkg2_t *pkg2, requirement_t *r);

static flag_t
web_path_req(FILE *out, pkg2_t *pkg2, index_t i_p);

static flag_t
knot_reqs_list(FILE *out, pkg2_t *pkg2, knot_t *k);


/*! writes all requirements concerning the current knot

\param out output file handle
\param pkg2 pkg2 record
\param k knot to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
 */
flag_t
web_requirements(FILE *out, pkg2_t *pkg2, knot_t *k)
{
  index_t r_i;
  index_t e_i;
  index_t q_i;
  index_t k_i;
  knot_t *knots;
  unsigned char buffer_name[BUFFER_LENGTH];

  if (k) {

    pkg2_all_out_flag(pkg2, false);

    /* start the table */
    fprintf(out,
	    "<table><tbody>"
	    );

    /* find all related knots of k for use as columns */
/*     petrinet_all_out_flag(pkg2->pn, false); */
    for (e_i=0; e_i < pkg2->pn->n_e; e_i++) { /* all edges */
      if (pkg2->pn->edge[e_i].valid) {
	if (k->type == state) {
	  if (pkg2->pn->edge[e_i].from == k) {
	    pkg2->pn->edge[e_i].to->out = true;
	  }
	}
	else {
	  if (pkg2->pn->edge[e_i].to == k) {
	    pkg2->pn->edge[e_i].from->out   = true;
	  }
	}
      }
    }

    /*! depending on mode_develop find the list requirements
     */
    pkg2_knot_requirement_out_flag(pkg2, k, true);

    fprintf(out,
	    "<tr>\n"
	    "<th>Anforderung</th>"
	    "<th>%s-typisch</th>",
	    (k->type == state) ? "modell" : "prozess");

    /* write all column headers of the table */
    for (k_i=0, knots = ((k->type == state) ? pkg2->pn->t : pkg2->pn->s);
	 knots[k_i].id;
	 k_i++) {
      if (knots[k_i].valid && knots[k_i].out) {
	fprintf(out, "<th>%s</th>\n", form_knot(buffer_name, pkg2->pn, &(knots[k_i])));
      }
    }

    fprintf(out, "</tr>\n");

    /* list all requirements as rows */
    for (r_i=0; r_i < pkg2->n_r; r_i++) {
      /* find the requirement w/ right id */
      if (pkg2->r[r_i].valid && pkg2->r[r_i].out) {
	fprintf(out, "<tr>\n");

	/* first column: requirement */
	fprintf(out,
		"<td>%s</td>\n",
		form_requirement(buffer_name, pkg2, &(pkg2->r[r_i])));

	/* second column: knot specific requirement */
	fprintf(out, "<td>%c</td>\n", cat4knot(pkg2, k, &(pkg2->r[r_i])));

	/* next columns:  */

	for (k_i=0, knots = ((k->type == state) ? pkg2->pn->t : pkg2->pn->s);
	     knots[k_i].id;
	     k_i++) {
	  if (knots[k_i].valid && knots[k_i].out) {
	    /* there is an edge between k and knots[k_i], now find all
	       reqs according to pkg2->r[r_i] */
	    fprintf(out,
		    "<td>%c/%c</td>\n",
		    cat4edge(pkg2, k, &(knots[k_i]), &(pkg2->r[r_i])),
		    cat4knot(pkg2, &(knots[k_i]), &(pkg2->r[r_i])));
	  }
	}
	fprintf(out, "</tr>\n");
      }
    }

    fprintf(out, "</tbody></table>\n\n");

#if 0
    fprintf(out,
	    "Anforderungen (Error - zurck in Ursprungsproze, W - kann im selben Proze behoben werden, I - berblick)"
	    "<ul>"
	    "<li>%stypisch<ul>\n", (k->type == state) ? "modell" : "prozess" );

    for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
      if (pkg2->req[q_i].knot == k) {
	for (r_i=0; r_i < pkg2->n_r; r_i++) {
	  /* find the requirement w/ right id */
	  if (pkg2->r[r_i].valid && xmlStrEqual(pkg2->req[q_i].ref, pkg2->r[r_i].id)) {
	    fprintf(out,
		    "<li>%s (%s)</li>\n",
/* 		    pkg2->r[r_i].id, */
/* 		    isolat1_str(buffer_name, pkg2->r[r_i].name, BUFFER_LENGTH), */
		    form_requirement(buffer_name, &(pkg2->r[r_i]), empty_str),
		    pkg2->req[q_i].cat
		    );
	  }
	}
      }
    }
    fprintf(out, "</ul></li>\n\n");


    fprintf(out,
	    "<li>%sspezifisch<ul>\n", (k->type == state) ? "prozess" : "modell" );

    /* list all requirements refered by edges with FROM=k 
     */
    for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
      if (pkg2->req[q_i].edge
	  && ((k->type == state && pkg2->req[q_i].edge->from == k)
	      || (k->type == transition && pkg2->req[q_i].edge->to == k))) {
	for (r_i=0; r_i < pkg2->n_r; r_i++) {
	  /* find the requirement w/ right id */
	  if (pkg2->r[r_i].valid && xmlStrEqual(pkg2->req[q_i].ref, pkg2->r[r_i].id)) {
	    fprintf(out,
		    "<li>%s",
		    form_requirement(buffer_name, &(pkg2->r[r_i]), empty_str));

	    fprintf(out,
		    " (%s &larr; %s)</li>\n",
		    pkg2->req[q_i].cat,
		    form_knot(buffer_name, pkg2->pn, ((k->type == state) ? pkg2->req[q_i].edge->to : pkg2->req[q_i].edge->from), pkg2->pn->doc_prefix)
		    );
	  }
	}
      }
    }
    fprintf(out, "</ul></ul>\n\n");


#endif
  }

  return true;
}
/* End of web_requirements() */


/*! writes all requirements into one table

\param out output file handle
\param pkg2 pkg2 record
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
 */
flag_t
web_requirements_table(FILE *out, pkg2_t *pkg2)
{

  index_t n;
  index_t i;
  index_t j;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\"><tbody>\n");

  fprintf( out, "<tr><th colspan=\"6\">%s</th></tr>\n", "Anforderungskatalog");

    fprintf( out,
	     "<tr>"
	     "<th>Nr.</th>\n"
	     "<th>Subnetz</th>\n"
	     "<th>Name</th>\n"
	     "<th>Beschreibung</th>\n"
	     "<th>Links</th>\n"
	     "<th>Prfungen</th>\n"
	     "</tr>\n");

  for (i=0, n=0; pkg2->r[i].id && i < pkg2->n_r; i++) {

    if (pkg2->r[i].valid) {
      fprintf( out, "<tr>");

      fprintf( out,
	       "<td>%i</td>\n",
	       ++n );

      fprintf( out,
	       "<td>%s</td>\n",
	       form_subnet(buffer_name, pkg2->pn, pkg2->r[i].sn));

      fprintf( out, "<td>%s</td>\n",
	       form_requirement(buffer_name, pkg2, &(pkg2->r[i])));

      fprintf( out, "<td>%s</td>\n",
	       isolat1_str(buffer_name, pkg2->r[i].description, BUFFER_LENGTH));

      fprintf( out, "<td>\n");
      web_htitem_list(out, pkg2->r[i].l);
      fprintf( out, "</td>\n");

      fprintf( out, "<td>\n");
      web_htitem_list(out, pkg2->r[i].check);
      fprintf( out, "</td>\n");

      fprintf( out, "</tr>\n");
    }
  }
  fprintf( out, "</tbody></table>\n");

  return true;
}
/* End of web_requirements_table() */


/*! writes all data of concerning a requirements into a card

\param out output file handle
\param pkg2 pkg2 record
\param r requirement to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
 */
flag_t
web_requirements_card(FILE *out, pkg2_t *pkg2, requirement_t *r)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  if (r->id && r->valid && r->sn->valid) {

    web_header(out, pkg2->pn);

    web_form_mode(out, pkg2);

    fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\" width=\"80%%\"><tbody>\n");
  
    fprintf( out,
	     "<tr><th colspan=\"1\">Anforderung: %s \n",
	     ((r->name && strlen((char *) r->name))
	      ? isolat1_str(buffer_name, r->name, BUFFER_LENGTH) : r->id));

    fprintf( out,
	     "(%s)</th></tr>\n",
	     ((r->sn->name && strlen((char *) r->sn->name)) ? isolat1_str(buffer_name, r->sn->name, BUFFER_LENGTH) : r->sn->id) );

    fprintf( out, "<tr><td>%s</td></tr>\n",
             isolat1_str(buffer_name, r->description, BUFFER_LENGTH));


    fprintf( out, "<tr><td>Prfungen\n");
    web_htitem_list(out, r->check);
    fprintf( out, "</td></tr>\n");

    if (mode_docs(pkg2)) {
      fprintf( out, "<tr><td>\n");
      web_card_doc(out, r->id, pkg2->pn->html_prefix);
      fprintf( out, "</td></tr>\n");
    }

    /* link list */
    if (mode_links(pkg2) && r->l && r->l->text) {
      fprintf( out, "<tr><td>Links\n");
      web_htitem_list(out, r->l);
      fprintf( out, "</td></tr>\n");
    }

#if 0
    if (mode_requirements(pkg2)) {
      fprintf( out, "<tr><td>\n");
      web_requirements(out, pkg2, k);
      fprintf( out, "</td></tr>\n");
    }
#endif

    fprintf( out, "</tbody></table>\n");

    web_footer(out, pkg2->pn);
  }

  return true;
}
/* End of web_requirements_card() */


flag_t
web_reqs_table(FILE *out, pkg2_t *pkg2)
{

  index_t i;
  index_t j;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\"><tbody>\n");

  fprintf( out, "<tr><th colspan=\"6\">%s</th></tr>\n", "Anforderungen");

  fprintf( out,
	   "<tr>"
	   "<th>Id Pruefung</th>\n"
	   "<th>Kategorie</th>\n"
	   "<th>Modell</th>\n"
	   "<th>Prozess</th>\n"
	   "<th>Rel. von</th>\n"
	   "<th>Rel. nach</th>\n"
	   "</tr>\n");

  for (i=0; pkg2->req[i].ref; i++) {
    fprintf( out, "<tr>");

    fprintf( out,
	     "<td>%s</td>\n",
	     pkg2->req[i].ref );

    fprintf( out,
	     "<td>%s</td>\n",
	     (pkg2->req[i].cat ? pkg2->req[i].cat : "-") );

    fprintf( out,
	     "<td>%s</td>\n",
	     ((pkg2->req[i].knot && pkg2->req[i].knot->type == state) ? pkg2->req[i].knot->id : "-") );

    fprintf( out,
	     "<td>%s</td>\n",
	     ((pkg2->req[i].knot && pkg2->req[i].knot->type == transition) ? pkg2->req[i].knot->id : "-") );

    fprintf( out,
	     "<td>%s</td>\n",
	     ((pkg2->req[i].edge && pkg2->req[i].edge->valid) ? pkg2->req[i].edge->from->id : "-") );


    fprintf( out,
	     "<td>%s</td>\n",
	     ((pkg2->req[i].edge && pkg2->req[i].edge->valid) ? pkg2->req[i].edge->to->id : "-") );


#if 0
    fprintf( out, "<td>%s</td>\n",
             isolat1_str(buffer_name, pkg2->req[i].name, BUFFER_LENGTH));

    fprintf( out, "<td>%s</td>\n",
             isolat1_str(buffer_name, pkg2->req[i].description, BUFFER_LENGTH));

    fprintf( out, "<td>\n");
    web_htitem_list(out, pkg2->req[i].l);
    fprintf( out, "</td>\n");

    fprintf( out, "<td>\n");
    web_htitem_list(out, pkg2->req[i].check);
#endif
    fprintf( out, "</td>\n");

    fprintf( out, "</tr>\n");
  }

  fprintf( out, "</tbody></table>\n");

  return true;
}
/* End of web_reqs_table() */


/*! writes all requirements concerning k

\param out output file handle
\param pkg2 pkg2 record
\param k knot to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
 */
flag_t
knot_reqs_list(FILE *out, pkg2_t *pkg2, knot_t *k)
{

  index_t q_i;
  index_t r_i;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf(out, "<ul>");

  for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
    if (pkg2->req[q_i].knot == k) {
      for (r_i=0; r_i < pkg2->n_r; r_i++) {
	/* find the requirement w/ right id */
	if (pkg2->r[r_i].valid && xmlStrEqual(pkg2->req[q_i].ref, pkg2->r[r_i].id)) {
	  fprintf(out,
		  "<li>%s (%s)</li>\n",
		  form_requirement(buffer_name, pkg2, &(pkg2->r[r_i])),
		  pkg2->req[q_i].cat
		  );
	}
      }
    }
  }
  fprintf(out, "</ul>\n");

  return true;
}
/* End of knot_reqs_list() */


/*! writes all requirements concerning k

\param out output file handle
\param pkg2 pkg2 record
\param k knot to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)
 */
flag_t
edge_reqs_list(FILE *out, pkg2_t *pkg2, edge_t *e)
{

  index_t q_i;
  index_t r_i;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf(out, "<ul>");

  for (q_i=0; pkg2->req[q_i].ref; q_i++) { /* all reqs */
    if (pkg2->req[q_i].edge == e) {
      for (r_i=0; r_i < pkg2->n_r; r_i++) {
	/* find the requirement w/ right id */
	if (pkg2->r[r_i].valid && xmlStrEqual(pkg2->req[q_i].ref, pkg2->r[r_i].id)) {
	  fprintf(out,
		  "<li>%s (%s)</li>\n",
		  form_requirement(buffer_name, pkg2, &(pkg2->r[r_i])),
		  pkg2->req[q_i].cat
		  );
	}
      }
    }
  }
  fprintf(out, "</ul>\n");

  return true;
}
/* End of edge_reqs_list() */


/*! writes a html table of the current path

derived from web_path()

\todo merge back with web_path() ?
 */
flag_t
web_path_req(FILE *out, pkg2_t *pkg2, index_t i_p)
{
  int i, j, m;
  knot_t *k;
  unsigned char buffer_name[BUFFER_LENGTH];

  fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\"><tbody>\n");

/*   fprintf( out, "<tr><th colspan=\"5\">%s %i</th></tr>\n", "Path", i_p); */
  
  for (i=0, m=0; i <= pkg2->pn->l_path; i++) {

    k = pkg2->pn->path[i];

#if 1
    /* insert description of relation */
    if (i>0) {
/*     if (i>0 && edgep(pkg2->pn, pkg2->pn->path[i-1],k)) { */
      edge_t *e = edge_ptr4id(pkg2->pn, pkg2->pn->path[i-1]->id, k->id);
      assert(e);

      if (e->description != empty_str)
	fprintf( out,
		 "<tr><td></td><td></td><td></td><td>%s</td></tr>\n",
 		 e->description);
    }
#endif

    fprintf( out, "<tr>\n");

    if (k->type == state) {
      fprintf( out,
	       "<td></td>"
	       "<td class=\"%s\">%s</td>\n",
	       "state",
	       form_knot(buffer_name, pkg2->pn, k));
    }
    else {
      m++;
      fprintf( out,
	       "<td>%i</td>"
	       "<td class=\"%s\">%s</td>\n",
	       m,
	       "transition",
	       form_knot(buffer_name, pkg2->pn, k));
    }

    fprintf( out,
	     "<td>%s</td>\n",
	     form_subnet(buffer_name, pkg2->pn, k->sn));

    fprintf( out, "<td>%s</td>\n",
	     isolat1_str(buffer_name, k->description, BUFFER_LENGTH));

#if 0  
    /* link list */
    fprintf( out, "<td>\n");
    web_htitem_list(out, k->l);
    fprintf( out, "</td>\n");
#endif

    if (mode_requirements(pkg2)) {
      fprintf( out, "<td>\n");
      knot_reqs_list(out, pkg2, k);
      fprintf( out, "</td>\n");
    }

    fprintf( out, "</tr>\n");

    if (mode_requirements(pkg2) && i < pkg2->pn->l_path) {
      fprintf( out, "<tr><td></td><td></td><td></td><td></td><td>\n");
      edge_reqs_list(out, pkg2, edge_ptr4id(pkg2->pn,k->id,pkg2->pn->path[i+1]->id));
      fprintf( out, "</td></tr>\n");
    }
  }
  fprintf( out, "</tbody><table>\n");

  return true;
}
/* End of web_path_req() */


/*! derived from web_pathnet()

*/
flag_t
web_pathnet_req(FILE *out, pkg2_t *pkg2)
{
  /* 
  */
  index_t i = 0;

  if (pkg2->pn->start && pkg2->pn->start->valid
      && pkg2->pn->target && pkg2->pn->target->valid) {

    do {
      if (path_forward(pkg2->pn)) {
	i++;
	if (i< 20) {
/* 	  web_path(out, pkg2->pn, doc_prefix); */
	  web_path_req(out, pkg2, i);
	}
      }
    } while (branch_backward(pkg2->pn));

    fprintf(out,
	    "<p>%i path between '%s' and '%s' found</p>",
	    pkg2->pn->n_paths, 
	    pkg2->pn->start->id, 
	    pkg2->pn->target->id);
  }

  return (pkg2->pn->n_paths > 0);
}
/* End of web_pathnet_req() */


/*! writes a simple cardfile as table for the knot 'k'

\param out output file handle
\param pkg2 pkg2 record
\param k knot to describe
\param doc_prefix URL prefix to documentation

\return nothing (but an error flag)

\todo use modules from web_knot_card()
*/
flag_t
web_knot_card_req(FILE *out, pkg2_t *pkg2, knot_t *k)
{
  index_t i;
  unsigned char buffer_name[BUFFER_LENGTH];

  if (k->id && k->valid && k->sn->valid) {

    web_header(out, pkg2->pn);

    web_form_mode(out, pkg2);

    fprintf( out, "<table border=\"0\" cellspacing=\"1\" cellpadding=\"10\" width=\"80%%\"><tbody>\n");
  
    fprintf( out,
	     "<tr><th colspan=\"1\">%s: %s ",
	     (k->type == state ? "Modell" : "Proze&szlig;"),
	     form_knot(buffer_name, pkg2->pn, k));

    fprintf( out,
	     "(%s)</th></tr>\n",
	     form_subnet(buffer_name, pkg2->pn, k->sn));

    fprintf( out,
	     "<tr><td>%s</td></tr>\n",
             isolat1_str(buffer_name, k->description, BUFFER_LENGTH));

    if (mode_images(pkg2)) {
      /* Graph cell with sensitive png */
      fprintf( out, "<tr><td align=\"center\">\n");
      web_card_img(out, k->id, pkg2->pn->doc_prefix);
      fprintf( out, "</td></tr>\n");
    }

    if (mode_docs(pkg2)) {
      /* Graph cell with sensitive png */
      fprintf( out, "<tr><td>\n");
      web_card_doc(out, k->id, pkg2->pn->html_prefix);
      fprintf( out, "</td></tr>\n");
    }

    /* link list */
    if (mode_links(pkg2) && k->l && k->l->text) {
      fprintf( out, "<tr><td>Links\n");
      web_htitem_list(out, k->l);
      fprintf( out, "</td></tr>\n");
    }

    if (mode_requirements(pkg2)) {
      fprintf( out, "<tr><td>\n");
      web_requirements(out, pkg2, k);
      fprintf( out, "</td></tr>\n");
    }

    fprintf( out, "</tbody><table>\n");

    web_footer(out, pkg2->pn);
  }

  return true;
}
/* End of web_knot_card_req() */


/*! web_knot_card_file() - 

*/
flag_t
web_knot_card_req_file(pkg2_t *pkg2, knot_t *k)
{

  assert(k);

  if (k->id && k->valid && k->sn->valid) {

    char card_name[100];
    FILE *filep_knot_card;

    assert(k->name);

    sprintf(card_name, "%s%s.html", pkg2->pn->doc_prefix, (char *) k->id);

    filep_knot_card = fopen(card_name, "w");

    printf_exit( (! filep_knot_card), 2, "Cant create '%s'", card_name );

    printf_debug("Write '%s'", card_name);

    web_knot_card_req(filep_knot_card, pkg2, k);

    fclose(filep_knot_card);

    return true;
  }

  return false;
} /* End of web_knot_card_req_file() */


flag_t
web_requirements_card_file(pkg2_t *pkg2, requirement_t *r)
{
  assert(r);

  if (r->id && r->valid && r->sn->valid) {

    char card_name[100];
    FILE *filep_requirement_card;

    assert(r->name);

    sprintf(card_name, "%s%s.html", pkg2->pn->doc_prefix, (char *) r->id);

    filep_requirement_card = fopen(card_name, "w");

    printf_exit( (! filep_requirement_card), 2, "Cant create '%s'", card_name );

    printf_debug("Write '%s'", card_name);

    web_requirements_card(filep_requirement_card, pkg2, r);

    fclose(filep_requirement_card);

    return true;
  }

  return false;
} /* End of web_requirements_card_file() */


unsigned char *
form_requirement(unsigned char *buffer, pkg2_t *pkg2, requirement_t *r)
{
  unsigned char buffer_name[BUFFER_LENGTH];

  assert(r);

  if (r->id) {
    sprintf( buffer, "<a href=\"%s%s.html\">%s</a>\n",
	     ((pkg2->pn->mode_run == cgi) ? pkg2->pn->doc_prefix : empty_str),
             r->id,
	     isolat1_str(buffer_name, ((r->name && xmlStrlen(r->name) > 2) ? r->name : r->id), BUFFER_LENGTH)
             );
  }

  return buffer;
}
/* Ende von form_requirement() */


void
web_form_mode(FILE *out, pkg2_t *pkg2)
{

  fprintf( out, 
	   "<form action=\"%s\" method=\"post\">\n"
	   "<p>"
	   "   <input type=\"hidden\" name=\"cfg_file\" value=\"%s%s\">\n",
	   pkg2->pn->executable,
	   pkg2->pn->filename_absolut, pkg2->pn->filename
	   );

  fprintf( out, 
	   "   <input type=\"submit\" name=\"mode\" value=\"Modelle\">\n"
	   "   <input type=\"submit\" name=\"mode\" value=\"Prozesse\">\n"
	   );

  if (mode_requirements(pkg2)) {
    fprintf( out, 
	     "   <input type=\"submit\" name=\"mode\" value=\"Anforderungen\">\n"
	     "   <input type=\"submit\" name=\"mode\" value=\"Zuordnung\">\n"
	     );
  }

  if (mode_images(pkg2)) {
    fprintf( out, 
	   "   <input type=\"submit\" name=\"mode\" value=\"Netze\">\n"
	     );
  }

  fprintf( out, 
	   "   <input type=\"submit\" name=\"mode\" value=\"Suche\">\n"
	   "   <input type=\"submit\" name=\"mode\" value=\"Hilfe\">\n"
	   "</p>"
	   "</form>"
    	   "<hr>\n"
	   );

}
/* End of write_form_mode() */


